#!/usr/bin/env python
"""
File       : writer.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Module for Writers used in multiple experiment-specific workflows.
"""

# System modules
import os
from typing import Optional

# Third party modules
import numpy as np
from pydantic import (
    conint,
    constr,
    model_validator,
)

# Local modules
from CHAP import Writer
from CHAP.pipeline import PipelineItem
from CHAP.writer import validate_writer_model


def validate_model(model):
    if model.filename is not None:
        validate_writer_model(model)
    return model


def write_matplotlibfigure(data, filename, savefig_kw, force_overwrite=False):
    """Write a Matplotlib figure to file.

    :param data: The figure to write to file
    :type data: matplotlib.figure.Figure
    :param filename: File name.
    :type filename: str
    :param savefig_kw: Keyword args to pass to
        matplotlib.figure.Figure.savefig.
    :type savefig_kw: dict, optional
    :param force_overwrite: Flag to allow data to be overwritten if it
        already exists, defaults to `False`.
    :type force_overwrite: bool, optional
    """
    # Third party modules
    from matplotlib.figure import Figure

    if not isinstance(data, Figure):
        raise TypeError('Cannot write object of type'
                        f'{type(data)} as a matplotlib Figure.')

    if os.path.isfile(filename) and not force_overwrite:
        raise FileExistsError(f'{filename} already exists')

    if savefig_kw is None:
        data.savefig(filename)
    else:
        data.savefig(filename, **savefig_kw)

def write_nexus(data, filename, force_overwrite=False):
    """Write a NeXus object to file.

    :param data: The data to write to file
    :type data: nexusformat.nexus.NXobject
    :param filename: File name.
    :type filename: str
    :param force_overwrite: Flag to allow data to be overwritten if it
        already exists, defaults to `False`.
    :type force_overwrite: bool, optional
    """
    # Third party modules
    from nexusformat.nexus import NXobject

    if not isinstance(data, NXobject):
        raise TypeError('Cannot write object of type'
                        f'{type(data).__name__} as a NeXus file.')

    mode = 'w' if force_overwrite else 'w-'
    data.save(filename, mode=mode)

def write_tif(data, filename, force_overwrite=False):
    """Write a tif image to file.

    :param data: The data to write to file
    :type data: numpy.ndarray
    :param filename: File name.
    :type filename: str
    :param force_overwrite: Flag to allow data to be overwritten if it
        already exists, defaults to `False`.
    :type force_overwrite: bool, optional
    """
    # Third party modules
    from imageio import imwrite

    data = np.asarray(data)
    if data.ndim != 2:
        raise TypeError('Cannot write object of type'
                        f'{type(data).__name__} as a tif file.')

    if os.path.isfile(filename) and not force_overwrite:
        raise FileExistsError(f'{filename} already exists')

    imwrite(filename, data)

def write_txt(data, filename, force_overwrite=False, append=False):
    """Write plain text to file.

    :param data: The data to write to file
    :type data: Union[str, list[str]]
    :param filename: File name.
    :type filename: str
    :param force_overwrite: Flag to allow data to be overwritten if it
        already exists, defaults to `False`.
    :type force_overwrite: bool, optional
    :param append: Flag to allow data to be appended to the file if it
        already exists, defaults to `False`.
    :type append: bool, optional
    """
    # Local modules
    from CHAP.utils.general import is_str_series

    if not isinstance(data, str) and not is_str_series(data, log=False):
        raise TypeError('input data must be a str or a tuple or list of str '
                        f'instead of {type(data)} ({data})')

    if not force_overwrite and not append and os.path.isfile(filename):
        raise FileExistsError(f'{filename} already exists')

    if append:
        with open(filename, 'a') as f:
            if isinstance(data, str):
                f.write(data)
            else:
                f.write('\n'.join(data))
    else:
        with open(filename, 'w') as f:
            if isinstance(data, str):
                f.write(data)
            else:
                f.write('\n'.join(data))

def write_yaml(data, filename, force_overwrite=False):
    """Write data to a YAML file.

    :param data: The data to write to file
    :type data: Union[dict, list]
    :param filename: File name.
    :type filename: str
    :param force_overwrite: Flag to allow data to be overwritten if it
        already exists, defaults to `False`.
    :type force_overwrite: bool, optional
    """
    # Third party modules
    import yaml

    if not isinstance(data, (dict, list)):
        raise TypeError('input data must be a dict or list.')

    if os.path.isfile(filename) and not force_overwrite:
        raise FileExistsError(f'{filename} already exists')

    with open(filename, 'w') as f:
        yaml.dump(data, f, sort_keys=False)

def write_filetree(data, outputdir='.', force_overwrite=False):
    """Write data to a file tree.

    :param data: The data to write to files
    :type data: nexusformat.nexus.NXobject
    :param outputdir: Output directory.
    :type filename: str, optional
    :param force_overwrite: Flag to allow data to be overwritten if it
        already exists, defaults to `False`.
    :type force_overwrite: bool, optional
    """
    # System modules
    from os import makedirs

    # Third party modules
    from nexusformat.nexus import (
        NXentry,
        NXgroup,
        NXobject,
        NXroot,
        NXsubentry,
    )

    if not isinstance(data, NXobject):
        raise TypeError('Cannot write object of type'
                        f'{type(data).__name__} as a file tree to disk.')

    # FIX: Right now this can bomb if MultiplePipelineItem
    # is called simultaneously from multiple nodes in MPI
    if not os.path.isdir(outputdir):
        makedirs(outputdir)

    for k, v in data.items():
        if isinstance(v, NXsubentry) and 'schema' in v.attrs:
            schema = v.attrs['schema']
            filename = os.path.join(outputdir, v.attrs['filename'])
            if schema == 'txt':
                write_txt(list(v.data), filename, force_overwrite)
            elif schema == 'json':
                write_txt(str(v.data), filename, force_overwrite)
            elif schema in ('yml', 'yaml'):
                from json import loads
                write_yaml(loads(v.data.nxdata), filename, force_overwrite)
            elif schema in ('tif',  'tiff'):
                write_tif(v.data, filename, force_overwrite)
            elif schema == 'h5':
                if any(isinstance(vv, NXsubentry) for vv in v.values()):
                    nxbase = NXroot()
                else:
                    nxbase = NXentry()
                for kk, vv in v.attrs.items():
                    if kk not in ('schema', 'filename'):
                        nxbase.attrs[kk] = vv
                for kk, vv in v.items():
                    if isinstance(vv, NXsubentry):
                        nxentry = NXentry()
                        nxbase[vv.nxname] = nxentry
                        for kkk, vvv in vv.items():
                            nxentry[kkk] = vvv
                    else:
                        nxbase[kk] = vv
                write_nexus(nxbase, filename, force_overwrite)
            else:
                raise TypeError(f'Files of type {schema} not yet implemented')
        elif isinstance(v, NXgroup):
            write_filetree(v, os.path.join(outputdir, k), force_overwrite)


class ExtractArchiveWriter(Writer):
    """Writer for tar files from binary data."""
    def write(self, data):
        """Take a .tar archive represented as bytes contained in `data`
        and write the extracted archive to files.

        :param data: The data to write to archive.
        :type data: list[PipelineData]
        :return: The achived data.
        :rtype: bytes
        """
        # System modules
        from io import BytesIO
        import tarfile

        data = self.unwrap_pipelinedata(data)[-1]

        with tarfile.open(fileobj=BytesIO(data)) as tar:
            tar.extractall(path=self.filename)

        return data


class FileTreeWriter(PipelineItem):
    """Writer for a file tree in NeXus format.

    :ivar force_overwrite: Flag to allow data to be overwritten if it
        already exists, defaults to `False`. Note that the existence
        of files prior to executing the pipeline is not possible since
        the filename(s) of the data are unknown during pipeline
        validation.
    :type force_overwrite: bool, optional
    """
    force_overwrite: Optional[bool] = False

    def write(self, data):
        """Write a NeXus format object contained in `data` to a 
        directory tree stuctured like the NeXus tree.

        :param data: The data to write to disk.
        :type data: list[PipelineData]
        :raises RuntimeError: If `filename` already exists and
            `force_overwrite` is `False`.
        :return: The data written to disk.
        :rtype: Union[nexusformat.nexus.NXroot,
            nexusformat.nexus.NXentry]
        """
        # Third party modules
        from nexusformat.nexus import (
            NXentry,
            NXroot,
        )

        data = self.unwrap_pipelinedata(data)[-1]
        nxentry = self.get_default_nxentry(data)
        write_filetree(nxentry, self.outputdir, self.force_overwrite)
        return data


class H5Writer(Writer):
    """Writer for H5 files from an nexusformat.nexus.NXdata object."""
    def write(self, data):
        """Write the NeXus object contained in `data` to hdf5 file.

        :param data: The data to write to file.
        :type data: list[PipelineData]
        :raises RuntimeError: If `filename` already exists and
            `force_overwrite` is `False`.
        :return: The data written to file.
        :rtype: nexusformat.nexus.NXobject
        """
        # Third party modules
        from h5py import File
        from nexusformat.nexus import NXdata

        data = self.unwrap_pipelinedata(data)[-1]
        if not isinstance(data, NXdata):
            raise ValueError('Invalid data parameter {(data)}')

        mode = 'w' if self.force_overwrite else 'w-'
        with File(self.filename, mode) as f:
            f[data.signal] = data.nxsignal
            for i, axes in enumerate(data.attrs['axes']):
                f[axes] = data[axes]
                f[data.signal].dims[i].label = \
                    f'{axes} ({data[axes].units})' \
                    if 'units' in data[axes].attrs else axes
                f[axes].make_scale(axes)
                f[data.signal].dims[i].attach_scale(f[axes])

        return data


class ImageWriter(PipelineItem):
    """Writer for saving image files.

    :ivar filename: Name of file to write to.
    :type filename: str, optional
    :ivar force_overwrite: Flag to allow data in `filename` to be
        overwritten if it already exists, defaults to `False`.
    :type force_overwrite: bool, optional
    :ivar remove: Flag to remove the dictionary from `data`,
        defaults to `False`.
    :type remove: bool, optional
    """
    filename: Optional[str] = None
    force_overwrite: Optional[bool] = False
    remove: Optional[bool] = True

    _validate_filename = model_validator(mode='after')(validate_model)

    def write(self, data):
        """Write the image(s) contained in `data` to file.

        :param data: The data to write to file.
        :type data: list[PipelineData]
        :raises RuntimeError: If a file already exists and
            `force_overwrite` is `False`.
        :return: The data written to disk.
        :rtype: list, dict, matplotlib.animation.FuncAnimation,
            numpy.ndarray
        """
        # System modules
        from io import BytesIO

        # Third party modules
        from matplotlib.animation import (
            ArtistAnimation,
            FuncAnimation,
        )

        # Local modules
        from CHAP.utils.general import save_iobuf_fig

        try:
            ddata = self.get_data(
                data, schema='common.write.ImageWriter', remove=self.remove)
        except ValueError:
            self.logger.warning(
                'Unable to find match with schema `common.write.ImageWriter`: '
                'return without writing')
            return None
        if isinstance(ddata, list):
            for (buf, fileformat), basename in ddata:
                self.filename = f'{basename}.{fileformat}'
                if not os.path.isabs(self.filename):
                    self.filename = os.path.join(self.outputdir, self.filename)
                if isinstance(buf, (ArtistAnimation, FuncAnimation)):
                    buf.save(self.filename)
                else:
                    save_iobuf_fig(
                        buf, self.filename,
                        force_overwrite=self.force_overwrite)
            return ddata

        if isinstance(ddata, dict):
            image_data = ddata['image_data']
            fileformat = ddata['fileformat']
        elif isinstance(ddata, tuple) and len(ddata) == 2:
            image_data = ddata[0]
            fileformat = ddata[1]
        else:
            image_data = ddata
        basename, ext = os.path.splitext(self.filename)
        if ext[1:] != fileformat:
            self.filename = f'{self.filename}.{fileformat}'
        if not os.path.isabs(self.filename):
            self.filename = os.path.join(self.outputdir, self.filename)
        if os.path.isfile(self.filename) and not self.force_overwrite:
            raise FileExistsError(f'{self.filename} already exists')
        if isinstance(image_data, BytesIO):
            save_iobuf_fig(
                image_data, self.filename,
                force_overwrite=self.force_overwrite)
        elif isinstance(image_data, np.ndarray):
            if image_data.ndim == 2:
                # Third party modules
                from imageio import imwrite

                imwrite(self.filename, image_data)
            elif image_data.ndim == 3:
                # Third party modules
                from tifffile import imwrite

                kwargs = {'bigtiff': True}
                imwrite(self.filename, image_data, **kwargs)
        elif isinstance(image_data, (ArtistAnimation, FuncAnimation)):
            image_data.save(self.filename)
        else:
            raise ValueError(f'Invalid image input type {type(image_data)}')
        return ddata


class MatplotlibAnimationWriter(Writer):
    """Writer for saving matplotlib animations.

    :ivar fps: Movie frame rate (frames per second), defaults to `1`.
    :type fps: int, optional
    """
    fps: Optional[conint(gt=0)] = 1

    def write(self, data):
        """Write the matplotlib.animation.ArtistAnimation object
        contained in `data` to file.

        :param data: The data to write to file.
        :type data: list[PipelineData]
        :return: The original animation.
        :rtype: matplotlib.animation.ArtistAnimation
        """
        data = self.unwrap_pipelinedata(data)[-1]
        extension = os.path.splitext(self.filename)[1]
        if not extension:
            data.save(f'{self.filename}.gif', fps=self.fps)
        elif extension == '.gif':
            data.save(self.filename, fps=self.fps)
        elif extension == '.mp4':
            data.save(self.filename, writer='ffmpeg', fps=self.fps)

        return data


class MatplotlibFigureWriter(Writer):
    """Writer for saving matplotlib figures to image files.

    :ivar savefig_kw: Keyword args to pass to
        matplotlib.figure.Figure.savefig.
    :type savefig_kw: dict, optional
    """
    savefig_kw: Optional[dict] = None

    def write(self, data):
        """Write the matplotlib.figure.Figure contained in `data` to
        file.

        :param data: The data to write to file.
        :type data: list[PipelineData]
        :raises RuntimeError: If `filename` already exists and
            `force_overwrite` is `False`.
        :return: The original figure object.
        :rtype: matplotlib.figure.Figure
        """
        data = self.unwrap_pipelinedata(data)[-1]
        write_matplotlibfigure(
            data, self.filename, self.savefig_kw, self.force_overwrite)

        return data


class NexusWriter(Writer):
    """Writer for NeXus files from `NXobject`-s.

    :ivar nxpath: Path to a specific location in the NeXus file tree
        to write to (ignored if `filename` does not yet exist).
    :type nxpath: str, optional
    """
    nxpath: Optional[constr(strip_whitespace=True, min_length=1)] = None

    def write(self, data):
        """Write the NeXus object contained in `data` to file.

        :param data: The data to write to file.
        :type data: list[PipelineData]
        :raises RuntimeError: If `filename` already exists and
            `force_overwrite` is `False`.
        :return: The data written to file.
        :rtype: nexusformat.nexus.NXobject
        """
        # Third party modules
        from nexusformat.nexus import (
            NXFile,
            NXentry,
            NXroot,
        )

        nxobject = self.get_data(data, remove=self.remove)

        nxname = nxobject.nxname
        if not os.path.isfile(self.filename) and self.nxpath is not None:
            self.logger.warning(
                f'{self.filename} does not yet exist, ignoring nxpath '
                f'argument ({self.nxpath})')
            self.nxpath = None
        if self.nxpath is None:
            nxclass = nxobject.nxclass
            if nxclass == 'NXroot':
                nxroot = nxobject
            elif nxclass == 'NXentry':
                nxroot = NXroot(nxobject)
                nxroot[nxname].set_default()
            else:
                nxroot = NXroot(NXentry(nxobject))
                if nxclass == 'NXdata':
                    nxroot.entry[nxname].set_default()
                nxroot.entry.set_default()
            write_nexus(nxroot, self.filename, self.force_overwrite)
        else:
            with NXFile(self.filename, 'rw') as nxfile:
                self.logger.debug(f'nxfile.mode = {nxfile.mode}')
                root = nxfile.readfile()
                if nxfile.get(self.nxpath) is None:
                    if nxfile.get(os.path.dirname(self.nxpath)) is not None:
                        self.nxpath, nxname = os.path.split(self.nxpath)
                    else:
                        self.logger.warning(
                            f'Path "{self.nxpath}" not present in {self.filename}. '
                            f'Using {root.NXentry[0].nxpath} instead.')
                        self.nxpath = root.NXentry[0].nxpath
                full_nxpath = os.path.join(self.nxpath, nxname)
                self.logger.debug(f'Full path for object to write: {full_nxpath}')
                if nxfile.get(full_nxpath) is not None:
                    self.logger.debug(
                        f'{full_nxpath} already exists in {self.filename}')
                    if self.force_overwrite:
                        self.logger.warning(
                            'Deleting existing NXobject at '
                            f'{full_nxpath, nxname} in {self.filename}')
                        del root[full_nxpath]
                try:
                    root[full_nxpath] = nxobject
                except Exception as exc:
                    raise exc
        return data


class NexusValuesWriter(Writer):
    """Writer for updating values in an existing NeXus file."""
    def write(self, data, filename, path_prefix=''):
        """Write new values specified in `data` to the exising NeXus
        file `filename`.

        :param data: List of dictionaries with the following entries
        -- `'path'` identifying the location of the `NXfield` object
        to which values will be written, `'data'` identifying the data
        to be written, and `'idx'` identifying the index / indicies of
        the `NXfield` to which the data will be written.
        :type data: list[PipelineData]
        :param filename: Name of an existing NeXus file to update.
        :type filename: str
        :param path_prefix: Prefix to use for all paths in input
            `data`, defaults to `''`.
        :type path_prefix: str, optional
        :returns: Original contenst of `data`.
        :rtype: list[dict[str, object]]
        """
        import os
        from nexusformat.nexus import NXFile

        data = self.unwrap_pipelinedata(data)[-1]

        for d in data:
            with NXFile(filename, 'a') as nxroot:
                self.nxs_writer(
                    nxroot=nxroot,
                    path=os.path.join(path_prefix, d['path']),
                    idx=d['idx'],
                    data=d['data']
                )

        return data

    def nxs_writer(self, nxroot, path, idx, data):
        """Write data to a specific NeXus file.

        This method writes `data` to a specified dataset within a NeXus
        file at the given index (`idx`). If the dataset does not
        exist, an error is raised. The method ensures that the shape
        of `data` matches the shape of the target slice before
        writing.

        :param nxroot: NeXus root object.
        :type nxroot: nexusformat.nexus.NXroot
        :param path: Path to the dataset inside the NeXus file.
        :type path: str
        :param idx: Index or slice where the data should be written.
        :type idx: tuple or int
        :param data: Data to be written to the specified slice in the
            dataset.
        :type data: numpy.ndarray or compatible array-like object
        :return: The written data.
        :rtype: numpy.ndarray or compatible array-like object
        :raises ValueError: If the specified dataset does not exist or
            if the shape of `data` does not match the target slice.
        """
        import numpy as np
        self.logger.info(f'Writing to {path} at {idx}')

        # Check if the dataset exists
        if path not in nxroot:
            raise ValueError(
                f'Dataset "{path}" does not exist in the NeXus file.')

        # Access the specified dataset
        dataset = nxroot[path]

        # Check that the slice shape matches the data shape
        data = np.asarray(data)
        if dataset[idx].shape != data.shape:
            raise ValueError(
                f'Data shape {data.shape} does not match the target slice '
                f'shape {dataset[idx].shape}.')

        # Write the data to the specified slice
        dataset[idx] = data
        self.logger.info(f'Data written to "{path}" at slice {idx}.')


class PyfaiResultsWriter(Writer):
    """Writer for results of one or more pyFAI integrations. Able to
    handle multiple output formats. Currently supported formats are:
    .npz, .nxs.
    """
    def write(self, data):
        """Save pyFAI integration results to a file. Format is
        determined automatically form the extension of `filename`.

        :param data: The data to write to file.
        :type data: Union[list[PipelineData],
            list[pyFAI.containers.IntegrateResult]]
        """
        from pyFAI.containers import Integrate1dResult, Integrate2dResult

        try:
            results = self.unwrap_pipelinedata(data)[0]
        except Exception:
            results = data
        if not isinstance(results, list):
            results = [results]
        if (not all([isinstance(r, Integrate1dResult) for r in results])
               and not all(
                   [isinstance(r, Integrate2dResult) for r in results])):
            raise Exception(
                'Bad input data: all items must have the same type -- either '
                'all pyFAI.containers.Integrate1dResult, or all '
                'pyFAI.containers.Integrate2dResult.')

        if os.path.isfile(self.filename):
            if self.force_overwrite:
                self.logger.warning(f'Removing existing file {self.filename}')
                os.remove(self.filename)
            else:
                raise Exception(f'{self.filename} already exists.')
        _, ext = os.path.splitext(self.filename)
        if ext.lower() == '.npz':
            self.write_npz(results, self.filename)
        elif ext.lower() == '.nxs':
            self.write_nxs(results, self.filename)
        else:
            raise Exception(f'Unsupported file format: {ext}')
        self.logger.info(f'Wrote to {self.filename}')
        return results

    def write_npz(self, results, filename):
        """Save `results` to the .npz file, `filename`."""

        data = {'radial': results[0].radial,
                'intensity': [r.intensity for r in results]}
        if hasattr(results[0], 'azimuthal'):
            # 2d results
            data['azimuthal'] = results[0].azimuthal
        if all([r.sigma for r in results]):
            # errors were included
            data['sigma'] = [r.sigma for r in results]

        np.savez(filename, **data)

    def write_nxs(self, results, filename):
        """Save `results` to the .nxs file, `filename`."""
        raise NotImplementedError


class TXTWriter(Writer):
    """Writer for plain text files from string or tuples or lists of
    strings.

    :ivar append: Flag to allow data in `filename` to be be appended,
        defaults to `False`.
    :type append: bool, optional
    """
    append: Optional[bool] = False

    def write(self, data):
        """Write a string or tuple or list of strings contained in 
        `data` to file.

        :param data: The data to write to file.
        :type data: list[PipelineData]
        :raises TypeError: If the object contained in `data` is not a
            `str`, `tuple[str]` or `list[str]`.
        :raises RuntimeError: If `filename` already exists and
            `force_overwrite` is `False`.
        :return: The data written to file.
        :rtype: str, tuple[str], list[str]
        """
        data = self.unwrap_pipelinedata(data)[-1]
        write_txt(data, self.filename, self.force_overwrite, self.append)

        return data


class YAMLWriter(Writer):
    """Writer for YAML files from `dict`-s."""
    def write(self, data):
        """Write the last matching dictionary contained in `data` to
        file (the schema mush match is a schema is provided).

        :param data: The data to write to file.
        :type data: list[PipelineData]
        :raises TypeError: If the object contained in `data` is not a
            `dict`.
        :raises RuntimeError: If `filename` already exists and
            `force_overwrite` is `False`.
        :return: The data pipeline.
        :rtype: list[PipelineData]
        """
        # Third party modules
        from pydantic import BaseModel

        # Local modules
        from CHAP.models import CHAPBaseModel

        def get_dict(data):
            if isinstance(data, dict):
                return data
            if isinstance(data, (BaseModel, CHAPBaseModel)):
                try:
                    return data.model_dump()
                except Exception:
                    pass
            return None

        schema = self.get_schema()
        yaml_dict = None
        if schema is not None:
            for i, d in reversed(list(enumerate(data))):
                if schema == d['schema']:
                    yaml_dict = get_dict(d['data'])
                    if yaml_dict is not None:
                        if self.remove:
                            data.pop(i)
                        break
        if yaml_dict is None:
            if schema is not None:
                self.logger.warning(
                    f'Unable to find match with schema {schema}: '
                    'try finding a dictionary without matching schema')
            for i, d in reversed(list(enumerate(data))):
                yaml_dict = get_dict(d['data'])
                if yaml_dict is not None:
                    if self.remove:
                        data.pop(i)
                    break
        write_yaml(yaml_dict, self.filename, self.force_overwrite)
        self.status = 'written' # Right now does nothing yet, but could
                                # add a sort of modification flag later
        return data


class ZarrValuesWriter(Writer):
    """Writer for updating values in arrays of an existing Zarr
    file.
    """
    def write(self, data, filename, path_prefix='', outputdir='.'):
        """Write `data` (from `saxswaxs.PyfaiIntegrationProcessor`) to
        the Zarr file `filename`.

        :param data: Results from `saxswaxs.PyfaiIntegrationProcessor`.
        :type data: list[PipelineData]
        :param filename: Name of Zarr file to which to write.
        :type filename: str
        :returns: Original results from
            `saxswaxs.PyfaiIntegrationProcessor`.
        :rtype: list[dict[str, object]]
        """
        import os
        import zarr

        # Open file in append mode to allow modifications
        zarrfile = zarr.open(filename, mode='a')

        # Get list of PyfaiIntegrationProcessor results to write
        data = self.unwrap_pipelinedata(data)[-1]
        for d in data:
            self.zarr_writer(
                zarrfile=zarrfile,
                path=os.path.join(path_prefix, d['path']),
                idx=d['idx'],
                data=d['data'])

        return data

    def zarr_writer(self, zarrfile, path, idx, data):
        """Write data to a specific Zarr dataset.

        This method writes `data` to a specified dataset within a Zarr
        file at the given index (`idx`). If the dataset does not
        exist, an error is raised. The method ensures that the shape
        of `data` matches the shape of the target slice before
        writing.

        :param zarrfile: Path to the Zarr file.
        :type zarrfile: zarr.core.group.Group
        :param path: Path to the dataset inside the Zarr file.
        :type path: str
        :param idx: Index or slice where the data should be written.
        :type idx: tuple or int
        :param data: Data to be written to the specified slice in the
            dataset.
        :type data: numpy.ndarray or compatible array-like object
        :return: The written data.
        :rtype: numpy.ndarray or compatible array-like object
        :raises ValueError: If the specified dataset does not exist or
            if the shape of `data` does not match the target slice.
        """
        self.logger.info(f'Writing to {path} at {idx}')
        # Check if the dataset exists
        if path not in zarrfile:
            raise ValueError(
                f'Dataset "{path}" does not exist in the Zarr file.')

        # Access the specified dataset
        dataset = zarrfile[path]

        # Check that the slice shape matches the data shape
        if dataset[idx].shape != data.shape and data.shape[0] == 1:
            data = np.squeeze(data, axis=0)
        if dataset[idx].shape != data.shape:
            raise ValueError(
                f'Data shape {data.shape} does not match the target slice '
                f'shape {dataset[idx].shape}.')

        # Write the data to the specified slice
        dataset[idx] = data
        self.logger.info(f'Data written to "{path}" at slice {idx}.')


class ZarrWriter(Writer):
    """Writer for zarr groups."""
    def write(self, data, filename, outputdir='.', force_overwrite=False):
        import asyncio
        from zarr.core.buffer import default_buffer_prototype
        from zarr.storage import LocalStore
        from zarr.abc.store import Store
        from zarr.core.group import AsyncGroup, Group

        async def copy_zarr_store_to_local_store(zarr_store, local_store):
            async for k in zarr_store.list():
                self.logger.info(f'Copying {k}')
                buf = await zarr_store.get(
                    k, prototype=default_buffer_prototype())
                await local_store.set(k, buf)

        zarr_obj = self.unwrap_pipelinedata(data)[-1]
        if isinstance(zarr_obj, Store):
            _zarr_store = zarr_obj
        elif isinstance(zarr_obj, (AsyncGroup, Group)):
            _zarr_store = zarr_obj.store
        else:
            raise TypeError(
                'Expected zarr.abc.store.Store, zarr.core.group.AsyncGroup, '
                f'or zarr.core.group.Group, got {type(zarr_obj)}'
            )
        _local_store = LocalStore(filename)
        asyncio.run(copy_zarr_store_to_local_store(
            _zarr_store, _local_store))
        return _local_store


if __name__ == '__main__':
    # Local modules
    from CHAP.writer import main

    main()



#!/usr/bin/env python
#-*- coding: utf-8 -*-
"""
File       : processor.py
Author     : Rolf Verberg
Description: Module for Processors used only by GIWAXS experiments
"""
# System modules
from json import loads
import os
from typing import Optional

# Third party modules
import numpy as np
from pydantic import (
    Field,
    PrivateAttr,
)

# Local modules
from CHAP.giwaxs.models import (
    GiwaxsConversionConfig,
    PyfaiIntegrationConfig,
)
from CHAP.pipeline import PipelineData
from CHAP.processor import Processor
from CHAP.utils.general import fig_to_iobuf


class GiwaxsConversionProcessor(Processor):
    """A processor for converting GIWAXS images from curved to
    rectangular coordinates.

    :ivar config: Initialization parameters for an instance of
        CHAP.giwaxs.models.GiwaxsConversionConfig
    :type config: dict, optional
    :ivar save_figures: Save .pngs of plots for checking inputs &
        outputs of this Processor, defaults to `False`.
    :type save_figures: bool, optional
    """
    pipeline_fields: dict = Field(
        default = {
            'config': 'giwaxs.models.GiwaxsConversionConfig'}, init_var=True)
    config: GiwaxsConversionConfig
    save_figures: Optional[bool] = True

    _figures: list = PrivateAttr(default=[])

    def process(self, data):
        """Process the GIWAXS input images & configuration and returns
        a map of the images in rectangular coordinates as a
        `nexusformat.nexus.NXroot` object.

        :param data: Results of `common.MapProcessor` containing the
            map of GIWAXS input images.
        :type data: list[PipelineData]
        :return: Converted GIWAXS images.
        :rtype: nexusformat.nexus.NXroot
        """
        # Third party modules
        from nexusformat.nexus import (
            NXentry,
            NXroot,
        )

        self._interactive = interactive
        self._save_figures = save_figures

        # Load the detector data
        try:
            nxobject = self.get_data(data)
            if isinstance(nxobject, NXroot):
                nxroot = nxobject
            elif isinstance(nxobject, NXentry):
                nxroot = NXroot()
                nxroot[nxobject.nxname] = nxobject
            else:
                raise ValueError(
                    f'Invalid nxobject in data pipeline ({type(nxobject)}')
        except Exception as exc:
            raise RuntimeError(
                'No valid detector data in input pipeline data') from exc

        nxroot = self.convert_q_rect(nxroot)
        if self._figures:
            return (
                nxroot,
                PipelineData(
                    name=self.__name__, data=self._figures,
                    schema='common.write.ImageWriter'))
        return nxroot


    def convert_q_rect(self, nxroot):
        """Return NXroot containing the converted GIWAXS images.

        :param nxroot: GIWAXS map with the raw detector data.
        :type nxroot: nexusformat.nexus.NXroot
        :return: Converted GIWAXS images.
        :rtype: nexusformat.nexus.NXroot
        """
        # Third party modules
        if self.interactive or self.save_figures:
            import matplotlib.pyplot as plt
        from nexusformat.nexus import (
            NXdata,
            NXfield,
            NXprocess,
        )

        # Local modules
        if self._save_figures:
            from CHAP.pipeline import PipelineData
            from CHAP.utils.general import fig_to_iobuf

        # Add the NXprocess object to the NXroot
        nxprocess = NXprocess()
        try:
            nxroot[f'{nxroot.default}_converted'] = nxprocess
        except Exception as exc:
            # Local imports
            from CHAP.utils.general import nxcopy

            # Copy nxroot if nxroot is read as read-only
            # FIX fix this in nexus reader
            nxroot = nxcopy(nxroot)
            nxroot[f'{nxroot.default}_converted'] = nxprocess
        nxprocess.conversion_config = self.config.model_dump_json()

        # Validate the azimuthal integrators and independent dimensions
        nxentry = nxroot[nxroot.default]
        nxdata = nxentry[nxentry.default]
        ais = self.config.azimuthal_integrators
        if len(ais) > 1:
            raise RuntimeError(
                'More than one azimuthal integrator not yet implemented')
        if ais[0].get_id() not in nxdata:
            raise RuntimeError('Unable to find detector data for '
                               f'{ais[0].get_id()} in {nxentry.tree}')
        if not isinstance(nxdata.attrs['axes'], str):
            raise RuntimeError(
                'More than one independent dimension not yet implemented')

        # Collect the raw giwaxs images
        scan_step_indices = self.config.scan_step_indices
        if scan_step_indices is None:
            thetas = nxdata[nxdata.attrs['axes']]
            giwaxs_data = nxdata[ais[0].get_id()]
        else:
            thetas = nxdata[nxdata.attrs['axes']][scan_step_indices]
            giwaxs_data = nxdata[ais[0].get_id()][scan_step_indices]
        self.logger.debug(f'giwaxs_data.shape: {giwaxs_data.shape}')
        effective_map_shape = giwaxs_data.shape[:-2]
        self.logger.debug(f'effective_map_shape: {effective_map_shape}')
        image_dims = giwaxs_data.shape[1:]
        self.logger.debug(f'image_dims: {image_dims}')

        # Get the components of q parallel and perpendicular to the
        # detector
        q_par, q_perp = self._calc_q_coords(giwaxs_data, thetas, ais[0].ai)

        # Get the range of the perpendicular component of q and that
        # of the parallel one at near grazing incidence as well as
        # the corresponding rectangular grid with the same dimensions
        # as the detector grid and converted to this grid from the
        # (q_par, q_perp)-grid
        # RV: For now use the same q-coords for all thetas, based on
        # the range for the first theta
        q_perp_min_index = np.argmin(np.abs(q_perp[:,0]))
        q_par_rect = np.linspace(
            q_par[q_perp_min_index,:].min(),
            q_par[q_perp_min_index,:].max(), image_dims[1])
        q_perp_rect = np.linspace(
            q_perp.min(), q_perp.max(), image_dims[0])
        figures = []
        giwaxs_data_rect = []
#        q_par_rect = []
#        q_perp_rect = []
        for i, theta in enumerate(thetas):
#            q_perp_min_index = np.argmin(np.abs(q_perp[i,:,0]))
#            q_par_rect.append(np.linspace(
#                q_par[i,q_perp_min_index,:].min(),
#                q_par[i,q_perp_min_index,:].max(), image_dims[1]))
#            q_perp_rect.append(np.linspace(
#                q_perp[i].min(), q_perp[i].max(), image_dims[0]))
#            giwaxs_data_rect.append(
#                GiwaxsConversionProcessor.curved_to_rect(
#                    giwaxs_data[i], q_par[i], q_perp[i], q_par_rect[i],
#                    q_perp_rect[i]))
            giwaxs_data_rect.append(
                GiwaxsConversionProcessor.curved_to_rect(
                    giwaxs_data[i], q_par, q_perp, q_par_rect,
                    q_perp_rect))

            if self.interactive or self.save_figures:
                vmax = giwaxs_data[i].max()/10
                fig, ax = plt.subplots(1,2, figsize=(10, 5))
                fig.suptitle(
                    f'Detector {ais[0].id}: 'r'$\theta$'f' = {theta}')
                # Right plot: imaged transformed to rectangular coords
                ax[1].imshow(
                    giwaxs_data_rect[i],
                    vmin=0, vmax=vmax,
                    origin='lower',
                    extent=(q_par_rect.min(), q_par_rect.max(),
                            q_perp_rect.min(), q_perp_rect.max()))
                ax[1].set_aspect('equal')
                ax[1].set_title('Transformed Image')
                ax[1].set_xlabel(r'q$_\parallel$'+' [\u212b$^{-1}$]')
                ax[1].set_ylabel(r'q$_\perp$'+' [\u212b$^{-1}$]')
                # Left plot: raw image
                im = ax[0].imshow(giwaxs_data[i], vmin=0, vmax=vmax)
                ax[0].set_aspect('equal')
                lhs = ax[0].get_position().extents
                rhs = ax[1].get_position().extents
                ax[0].set_position(
                    (lhs[0], rhs[1], rhs[2] - rhs[0], rhs[3] - rhs[1]))
                ax[0].set_title('Raw Image')
                ax[0].set_xlabel('column index')
                ax[0].set_ylabel('row index')
                fig.subplots_adjust(right=0.85)
                cbar_ax = fig.add_axes([0.9, 0.15, 0.025, 0.7])
                fig.colorbar(im, cax=cbar_ax)
                fig.suptitle(f'theta: {theta:.4f}')
                if self.save_figures:
                    if scan_step_indices is None:
                        basename = 'converted'
                    else:
                        basename = f'converted_{scan_step_indices[i]}'
                    self._figures.append(
                        (fig_to_iobuf(fig), f'{basename}_{i}'))
                if self.interactive:
                    plt.show()
                plt.close()

        # Create an animation of the figures
        animation = []
        if self._save_figures and not config.skip_animation:
            # Third party modules
            from matplotlib.animation import ArtistAnimation

            frames = []
            for (buf, _), _ in figures:
                buf.seek(0)
                frame = plt.imread(buf)
                im = plt.imshow(frame, animated=True)
                if not i:
                    plt.imshow(frame)
                frames.append([im])

            ani = ArtistAnimation(
                 plt.gcf(), frames, interval=1000, blit=False, repeat=False)
            animation.append(((ani, 'gif'), f'{ais[0].id}_converted'))

        # Create the NXdata object with the converted images
        nxprocess.data = NXdata(
            NXfield(
                np.asarray(giwaxs_data_rect), f'{ais[0].id}_converted'),
            (thetas,
             NXfield(
                 q_perp_rect, 'q_perp_rect',
                 attrs={'units': '\u212b$^{-1}$'}),
             NXfield(
                 q_par_rect, 'q_par_rect',
                 attrs={'units': '\u212b$^{-1}$'})))
        nxprocess.attrs['default'] = 'data'

        if self._save_figures:
            if animation:
                return (
                    nxroot,
                    PipelineData(
                        name=self.__name__, data=figures,
                        schema='common.write.ImageWriter'),
                    PipelineData(
                        name=self.__name__, data=animation,
                        schema='common.write.ImageWriter'))
            return (
                nxroot,
                PipelineData(
                    name=self.__name__, data=figures,
                    schema='common.write.ImageWriter'))
        return nxroot

    @staticmethod
    def curved_to_rect(
        data_curved, q_par, q_perp, q_par_rect, q_perp_rect,
        return_maps=False, normalize=True):
        """
        ::

            data_rect = curved_to_rect(...):
                distributes counts from a curvilinear grid
                (data_curved), e.g. x-ray data collected in angular
                space, into a rectilinear grid (reciprocal space). 
        
            data_rect, norm, xmap, ymap, xwid, ywid =
                    curved_to_rect(..., return_maps=True):
                distributes counts from a curvilinear grid
                (data_curved), e.g. x-ray data collected in angular
                space, into a rectilinear grid (reciprocal space). 
 
            q_par, q_perp, and data_curved are M x N following the
                normal convention where the the first & second index
                corrspond to the vertical (y) and horizontal (x)
                locations of the scattering pattern.
            q_par, q_perp represent the q coordinates of the center of
                pixels whose intensities are stored in data_curved.
                Reiterating the convention above, q_par and q_perp vary
                primarilly along the 2nd and 1st index, respectively.
            q_par_rect and q_perp_rect are evenly-spaced, monotonically
                increasing, arrays determining the new grid.
           
            data_rect : the new matrix with intensity from data_curved
                        disctributed into a regular grid defined by
                        q_par_rect, q_perp_rect.
            norm : a matrix with the same shape of data_rect
                   representing the area of the pixel in the original
                   angular units. It should be used to normalize the
                   resulting array as norm_z = data_rect / norm.

            Algorithm:
               Step 1 : Compute xmap, ymap, which containt the values
                        of q_par and q_perp, but represented in pixel
                        units of the target coordinates q_par_rect,
                        q_perp_rect.
                        In other words, xmap(i,j) = 3.4 means that
                        q_par(i,j) lands 2/5 of the q_distance between
                        q_par_rect(3) and q_par_rect(4). Intensity in
                        qpar(i,j) should thus be distributed in a 2:3
                        ratio among neighboring mini-columns of pixels
                        3 and 4.
               Step 2 : Use the procedure described by Barna et al
                        (RSI v.70, p. 2927, 1999) to distribute
                        intensity from each source pixel i,j into each
                        of 9 destination pixels around the xmap(i,j)
                        and ymap(i,j). Keep track of how many source
                        "pixels" are placed into each bin in the
                        variable, "norm". Note also that if
                        xmap(i,j)-floor(xmap(i,j)) > 0.5,
                        the "center" pixel of the 9 destination pixels
                        is floor(xmap+0.5).
               (Outside this function): The normalized intensity in
                   each new pixel can be obtained as
                   I = data_rect./norm, but with the caveat that zero
                   values of "norm" should be changed to ones first,
                   norm(data_rect == 0) = 1.0.

            Example Usage: 
                1. Compute the values of q_par and q_perp for each
                   pixel in the image z (according to scattering
                   geometry).
                2. Set or determing a good target grid, e.g.:        
                   min_qpar, max_qpar = q_par.mix(), q_par.max()
                   min_qperp, max_qperp = q_perp.mix(), q_perp.max()
                   q_par_rect, q_par_step = np.linspace(min_qpar ,
                       max_qpar, image_dim[1], retstep=True)
                   q_perp_rect, q_perp_step = np.linspace(min_qperp,
                       max_qperp, image_dim[0], retstep=True)
               3. data_rect = curved_to_rect(data_curved, q_par,
                      q_perp, q_par_rect, q_perp_rect)
               4. plt.imshow(data_rect, extent = [
                      q_par_rect[0], q_par_rect[-1],
                      q_perp_rect[-1], q_perp_rect[0]])
                  xlabel(r'q$_\parallel$'' [\u212b$^{-1}$]')
                  ylabel(r'q$_\perp$'' [\u212b$^{-1}$]')
        """
        out_width, out_height = q_par_rect.size, q_perp_rect.size

        # Check correct dimensionality
        dims = data_curved.shape
        assert q_par.shape == dims and q_perp.shape == dims

        data_rect = np.zeros((out_height, out_width))
        norm = np.zeros_like(data_rect)

        rect_width  = q_par_rect[1] - q_par_rect[0]
        rect_height = q_perp_rect[1]- q_perp_rect[0]
        q_par_rect_shift = q_par_rect - rect_width/2.0
        q_perp_rect_shift = q_perp_rect - rect_height/2.0

        # Precompute source pixels that are outside the target area
        out_of_bounds = (
            (q_par < q_par_rect_shift[0])
            | (q_par > q_par_rect_shift[-1] + rect_width)
            | (q_perp < q_perp_rect_shift[0])
            | (q_perp > q_perp_rect_shift[-1] + rect_height))

        # Vectorize the search for where q_par[i, j] and q_perp[i,j]
        # fall on the grid formed by q_par_rect and q_perp_rect
        #
        # 1. Expand q_par_rect_shift (a vector) such that
        #    q_par_rect_shift_cube[i. j, :] is identical to
        #    q_par_rect_shift, and is a rising sequence of values of
        #    qpar
        # 2. Expand q_par such that qpar_cube[i, j, :] all correspond
        #    to the value q_par[i, j].
        # - Note that I found tile first and used that in once case but
        #   not the other. I think broadcast_to should likely be used
        #   for both. But in both cases, it seemed to be easiest or
        #   only possible if the extra dimensions were leading dims,
        #   not trailing. That is the reason for the use of transpose
        #   in qpar_cube.
        q_par_rect_shift_cube = np.tile(q_par_rect_shift, q_par.shape + (1,))
        qpar_cube = np.transpose(np.broadcast_to(
            q_par, ((len(q_par_rect_shift),) + q_par.shape)), (1,2,0))
        q_perp_rect_shift_cube = np.tile(
            q_perp_rect_shift, q_perp.shape + (1,))
        qperp_cube = np.transpose(np.broadcast_to(
            q_perp, ((len(q_perp_rect_shift),) + q_perp.shape)), (1,2,0))

        # We want the index of the highest q_par_rect_shift that is
        # still below qpar, whereas the argmax # operation yields the
        # first q_par_rect_shift that is above qpar, We subtract 1 to
        # take care of this and then correct for any negative indices
        # to 0.
        highpx_x = np.argmax(qpar_cube < q_par_rect_shift_cube, axis=2) - 1
        highpx_y = np.argmax(qperp_cube < q_perp_rect_shift_cube, axis=2) - 1
        highpx_x[highpx_x < 0] = 0
        highpx_y[highpx_y < 0] = 0

        # Compute xmap and ymap
        xmap = np.where(
            out_of_bounds, np.nan,
            highpx_x - 0.5 + (q_par - q_par_rect_shift[highpx_x]) / rect_width)
        ymap = np.where(
            out_of_bounds, np.nan,
            highpx_y - 0.5
                + (q_perp - q_perp_rect_shift[highpx_y]) / rect_height)

        # Optionally, print out-of-bounds pixels
        if np.any(out_of_bounds):
            print(f'Warning: Found {out_of_bounds.sum()} source pixels that '
                  'are outside the target bounding box')
            # out_of_bounds_indices = np.transpose(np.where(out_of_bounds))
            # for i, j in out_of_bounds_indices:
            #     print(f'pixel {i}, {j} is out of bounds...skip')

        x1 = np.floor(xmap + 0.5).astype(int)
        y1 = np.floor(ymap + 0.5).astype(int)

        # Compute the effective size of each source pixel (for
        # comparison with target pixels)
        xwid = np.abs(np.diff(xmap, axis=1))
        ywid = np.abs(np.diff(ymap, axis=0))

        # Prepend xwid, ywid with their first column, row
        # (respectively) to match shape with xmap, ymap
        xwid = np.insert(xwid, 0, xwid[:,0], axis=1)
        ywid = np.insert(ywid, 0, ywid[0,:], axis=0)

        # Compute mapping of source pixel to up to 9 closest pixels,
        # after Barna (1999)
        col = np.zeros((3,)+xmap.shape)
        row = np.zeros((3,)+xmap.shape)
        col[0,:,:] = np.where(
            0.5 - (xmap - x1 + 0.5)/xwid > 0.0,
            0.5 - (xmap - x1 + 0.5)/xwid, 0.0)
        col[2,:,:] = np.where(
            0.5 + (xmap - x1 - 0.5)/xwid > 0.0,
            0.5 + (xmap - x1 - 0.5)/xwid, 0.0)
        col[1,:,:] = 1.0 - col[0,:,:] - col[2,:,:]
        row[0,:,:] = np.where(
            0.5 - (ymap - y1 + 0.5)/ywid > 0.0,
            0.5 - (ymap - y1 + 0.5)/ywid , 0.0)
        row[2,:,:] = np.where(
            0.5 + (ymap - y1 - 0.5)/ywid > 0.0,
            0.5 + (ymap - y1 - 0.5)/ywid, 0.0)
        row[1,:,:] = 1.0 - row[0,:,:] - row[2,:,:]

        for k in (-1, 0, 1):
            for m in (-1, 0, 1):
                source_indices = ((x1+k > -1) & (x1+k < out_width) &
                                  (y1+m > -1) & (y1+m < out_height))
                x1_sub = x1[source_indices]+k
                y1_sub = y1[source_indices]+m

                np.add.at(data_rect, (y1_sub, x1_sub),
                          data_curved[source_indices] *
                          col[k+1, source_indices] *
                          row[m+1, source_indices])
                np.add.at(norm, (y1_sub, x1_sub),
                          col[k+1, source_indices] * row[m+1, source_indices])
                # The following fails because although
                # [y1_sub, x1_sub] can refer to the same location more
                # than once, the "+=" operation acts on the original
                # value and does not have knowledge of incremental
                # changes.
                # data_rect[y1_sub, x1_sub] += (data_curved[source_indices]
                #   * col[k+1, source_indices] * row[m+1, source_indices])
                # norm[y1_sub, x1_sub] += \
                #     col[k+1, source_indices] * row[m+1, source_indices]

        if normalize:
            norm[norm == 0] = 1.0
            data_rect /= norm

        if return_maps:
            return data_rect, norm, xmap, ymap, xwid, ywid
        return data_rect

    def _calc_q_coords(self, images, thetas, ai):
        """Return a 3D arrays representing the perpendicular and
        parallel components of q relative to the detector surface
        for each pixel in an image for each theta.

        :param images: GIWAXS images.
        :type images: numpy.ndarray
        :param thetas: Image theta values.
        :type thetas: numpy.ndarray
        :param ai: Azimuthal integrator.
        :type ai: pyFAI.azimuthalIntegrator.AzimuthalIntegrator
        :return: Perpendicular and parallel components of q relative
            to the detector surface.
        :rtype: numpy.ndarray, numpy.ndarray
        """
        # Load the PONI file info:
        # PONI coordinates relative to the left bottom detector corner
        # viewed along the beam with the "1" and "2" directions along
        # the detector rows and columns, respectively
        assert ai.get_shape() == images.shape[1:]
        image_dim = ai.get_shape()
        sample_to_detector = ai.dist*1000     # Sample to detector  (mm)
        pixel_size = round(ai.pixel1*1000, 3) # Pixel size (mm)
        poni1 = ai.poni1*1000 # Point of normal incidence 1 (mm)
        poni2 = ai.poni2*1000 # Point of normal incidence 2 (mm)
        rot1 = ai.rot1        # Rotational angle 1 (rad)
        rot2 = ai.rot2        # Rotational angle 2 (rad)
        xray_wavevector = 2.e-10*np.pi / ai.wavelength

        # Pixel locations relative to where the incident beam
        # intersects the detector in the GIWAXS coordinates frame
        pixel_vert_position = (poni1 + sample_to_detector*np.tan(rot2)
                               - pixel_size*np.arange(image_dim[0]))
        pixel_hor_position = (pixel_size*np.arange(image_dim[1])
                             - poni2 + sample_to_detector*np.tan(rot1))

        # Deflection angles relative to the incident beam at each
        # pixel location
        delta = np.tile(
            np.arctan(pixel_vert_position/sample_to_detector),
            (image_dim[1],1)).T
        nu = np.tile(
            np.arctan(pixel_hor_position/sample_to_detector),
            (image_dim[0],1))
        sign_nu = 2*(nu>=0)-1

        # Calculate q_par, q_perp
        # RV: For now use the same q-coords for all thetas, based on
        # the range for the first theta
#        q_par = []
#        q_perp = []
#        for theta in thetas:
#            alpha = np.deg2rad(theta)
#            beta = delta - alpha;
#            cosnu = np.cos(nu);
#            cosb = np.cos(beta);
#            cosa = np.cos(alpha);
#            sina = np.sin(alpha);
#            q_par.append(sign_nu * xray_wavevector * np.sqrt(
#                cosa*cosa + cosb*cosb - 2*cosa*cosb*cosnu))
#            q_perp.append(xray_wavevector*(sina + np.sin(beta)))
        alpha = np.deg2rad(thetas[0])
        beta = delta - alpha
        cosnu = np.cos(nu)
        cosb = np.cos(beta)
        cosa = np.cos(alpha)
        sina = np.sin(alpha)
        q_par = sign_nu * xray_wavevector * np.sqrt(
            cosa*cosa + cosb*cosb - 2*cosa*cosb*cosnu)
        q_perp = xray_wavevector*(sina + np.sin(beta))

        return q_par, q_perp


class PyfaiIntegrationProcessor(Processor):
    """A processor for azimuthally integrating images.

    :ivar config: Initialization parameters for an instance of
        CHAP.giwaxs.models.GiwaxsConversionConfig
    :type config: dict, optional
    """
    pipeline_fields: dict = Field(
        default = {
            'config': 'giwaxs.models.PyfaiIntegrationConfig'}, init_var=True)
    config: PyfaiIntegrationConfig

    def process(self, data):
        """Process the input images & configuration and return a map of
        the azimuthally integrated images.

        :param data: Results of `common.MapProcessor` or other suitable
            preprocessor of the raw detector data containing the map of
            input images.
        :type data: list[PipelineData]
        :return: Integrated images.
        :rtype: nexusformat.nexus.NXroot
        """
        # Third party modules
        import fabio
        from nexusformat.nexus import (
            NXdata,
            NXentry,
            NXfield,
            NXprocess,
            NXroot,
            nxsetconfig,
        )
        from pyFAI.gui.utils.units import Unit

        # Local imports
        from CHAP.common.map_utils import get_axes
        from CHAP.utils.general import nxcopy

        nxsetconfig(memory=100000)

        # Load the detector data
        try:
            nxobject = self.get_data(data)
            if isinstance(nxobject, NXroot):
                nxroot = nxobject
            elif isinstance(nxobject, NXentry):
                nxroot = NXroot()
                nxroot[nxobject.nxname] = nxobject
                nxobject.set_default()
            else:
                raise ValueError(
                    f'Invalid nxobject in data pipeline ({type(nxobject)}')
        except Exception as exc:
            raise RuntimeError(
                'No valid detector data in input pipeline data') from exc

        # Validate the azimuthal integrator configuration and check
        # against the input data (availability and shape)
        data = {}
        try:
            nxprocess_converted = nxroot[f'{nxroot.default}_converted']
            conversion_config = loads(
                str(nxprocess_converted.conversion_config))
            converted_ais = conversion_config['azimuthal_integrators']
            if len(converted_ais) > 1:
                raise RuntimeError(
                    'More than one detector not yet implemented')
            if self.config.azimuthal_integrators is None:
                # Local modules
                from CHAP.common.models.integration import (
                    AzimuthalIntegratorConfig,
                )

                ais = [AzimuthalIntegratorConfig(**converted_ais[0])]
            else:
                converted_ids = [ai['id'] for ai in converted_ais]
                skipped_detectors = []
                ais = []
                for ai in self.config.azimuthal_integrators:
                    if ai.get_id() in converted_ids:
                        ais.append(ai)
                    else:
                        skipped_detectors.append(ai.get_id())
                if skipped_detectors:
                    self.logger.warning(
                        f'Skipping detector(s) {skipped_detectors} '
                        '(no converted data)')
                if not ais:
                    raise RuntimeError(
                        'No matching azimuthal integrators found')
            nxdata = nxprocess_converted.data
            axes = nxdata.attrs['axes']
            if len(nxdata.attrs['axes']) != 3:
                raise RuntimeError('More than one independent dimension '
                                   'not yet implemented')
            axes = axes[0]
            independent_dims[ais[0].get_id()] = nxcopy(nxdata[axes])
            data[ais[0].get_id()] = np.flip(nxdata.converted.nxdata, axis=1)
        except Exception as exc:
            experiment_type = loads(
                str(nxroot[nxroot.default].map_config))['experiment_type']
            if experiment_type == 'GIWAXS':
                self.logger.warning(
                    'No converted data found, use raw data for integration')
            nxentry = nxroot[nxroot.default]
            detector_ids = [
                #str(id, 'utf-8') for id in nxentry.detector_ids.nxdata]
                str(id) for id in nxentry.detector_ids.nxdata]
            if len(detector_ids) > 1:
                raise RuntimeError(
                    'More than one detector not yet implemented') from exc
            if self.config.azimuthal_integrators is None:
                raise ValueError('Missing azimuthal_integrators parameter in '
                                 f'PyfaiIntegrationProcessor.config '
                                 f'({self.config})') from exc
            nxdata = nxentry[nxentry.default]
            skipped_detectors = []
            ais = []
            for ai in self.config.azimuthal_integrators:
                if ai.get_id() in nxdata:
                    if nxdata[ai.get_id()].ndim != 3:
                        raise RuntimeError(
                            'Inconsistent raw data dimension '
                            f'{nxdata[ai.get_id()].ndim}') from exc
                    ais.append(ai)
                    data[ai.id] = nxdata[ai.id].nxdata
                else:
                    skipped_detectors.append(ai.get_id())
            if skipped_detectors:
                self.logger.warning('Skipping detector(s) '
                                    f'{skipped_detectors} (no raw data)')
            if not ais:
                raise RuntimeError(
                    'No matching raw detector data found') from exc
            if 'unstructured_axes' in nxdata.attrs:
                axes = nxdata.attrs['unstructured_axes']
                independent_dims[ais[0].get_id()] = [
                    nxcopy(nxdata[a]) for a in axes]
            elif 'axes' in nxdata.attrs:
                axes = nxdata.attrs['axes']
                independent_dims[ais[0].get_id()] = nxcopy(nxdata[axes])
            else:
                self.logger.warning('Unable to find independent_dimensions')
            data[ais[0].get_id()] = nxdata[ais[0].get_id()]

        # Select the images to integrate
        if False and self.config.scan_step_indices is not None:
            #FIX
            independent_dims = independent_dims[self.config.scan_step_indices]
            data = data[self.config.scan_step_indices]
        self.logger.debug(
            f'data shape(s): {[(k, v.shape) for k, v in data.items()]}')
        if self.config.sum_axes:
            data = {k:np.sum(v.nxdata, axis=0)[None,:,:]
                    for k, v in data.items()}
            self.logger.debug('data shape(s) after summing: '
                              f'{[(k, v.shape) for k, v in data.items()]}')

        # Read the mask(s)
        # FIX read at validation, like the poni file
        masks = {}
        for ai in ais:
            self.logger.debug(f'Reading {ai.mask_file}')
            try:
                with fabio.open(ai.mask_file) as f:
                    mask = f.data
                    self.logger.debug(
                        f'mask shape for {ai.get_id()}: {mask.shape}')
                    masks[ai.get_id()] = mask
            except Exception:
                self.logger.debug('No mask file found for {ai.get_id()}')
        if not masks:
            masks = None

        # Perform integration(s)
        ais_pyfai = {ai.get_id(): ai.ai for ai in ais}
        for integration in self.config.integrations:

            # Add a NXprocess object(s) to the NXroot
            nxprocess = NXprocess()
            try:
                nxroot[f'{nxroot.default}_{integration.name}'] = nxprocess
            except Exception:
                # Copy nxroot if nxroot is read as read-only
                nxroot = nxcopy(nxroot)
                nxroot[f'{nxroot.default}_{integration.name}'] = nxprocess
            nxprocess.integration_config = integration.model_dump_json()
            nxprocess.azimuthal_integrators = [
                ai.model_dump_json() for ai in ais]

            # Integrate the data
            results = integration.integrate(ais_pyfai, data, masks)

            # Create the NXdata object with the integrated data
            intensities = results['intensities']
            if self.config.sum_axes:
                coords = []
            elif isinstance(axes, str):
                coords = [
                    v for k, v in independent_dims.items() if k in ais_pyfai]
            else:
                coords = [i for k, v in independent_dims.items()
                          for i in v if k in ais_pyfai]
            if ('azimuthal' in results
                    and results['azimuthal']['unit'] == 'chi_deg'):
                chi = results['azimuthal']['coords']
                if integration.right_handed:
                    chi = -np.flip(chi)
                    intensities = np.flip(intensities, (len(coords)))
                coords.append(NXfield(chi, 'chi', attrs={'units': 'deg'}))
            if results['radial']['unit'] == 'q_A^-1':
                unit = Unit.INV_ANGSTROM.symbol
                coords.append(
                    NXfield(
                        results['radial']['coords'], 'q',
                        attrs={'units': unit}))
            else:
                coords.append(
                    NXfield(
                        results['radial']['coords'], 'r'))#,
#                        attrs={'units': '\u212b'}))
                self.logger.warning(
                    f'Unknown radial unit: {results["radial"]["unit"]}')
            nxdata = NXdata(NXfield(intensities, 'integrated'), tuple(coords))
            if not isinstance(axes, str):
                nxdata.attrs['unstructured_axes'] = nxdata.attrs['axes'][:-1]
                del nxdata.attrs['axes']
            nxprocess.data = nxdata
            nxprocess.default = 'data'

        self.config.azimuthal_integrators = ais

        return nxroot


if __name__ == '__main__':
    # Local modules
    from CHAP.processor import main

    main()



"""pyFAI integration related Pydantic model classes."""

# System modules
from copy import deepcopy
import os
from typing import (
    Literal,
    Optional,
    Union,
)

# Third party modules
import numpy as np
from pydantic import (
    FilePath,
    PrivateAttr,
    confloat,
    conint,
    conlist,
    constr,
    field_validator,
    model_validator,
)
from pyFAI.integrator.azimuthal import AzimuthalIntegrator

# Local modules
from CHAP import CHAPBaseModel
from CHAP.common.models.map import Detector


class AzimuthalIntegratorConfig(Detector, CHAPBaseModel):
    """Azimuthal integrator configuration class to represent a single
    detector used in the experiment.

    :param mask_file: Path to the mask file.
    :type mask_file: FilePath, optional
    :param poni_file: Path to the PONI file, specify either `poni_file`
        or `params`, not both.
    :type poni_file: FilePath, optional
    :param params: Azimuthal integrator configuration parameters,
        specify either `poni_file` or `params`, not both.
    :type params: dict, optional
    """
    mask_file: Optional[FilePath] = None
    params: Optional[dict] = None
    poni_file: Optional[FilePath] = None

    _ai: AzimuthalIntegrator = PrivateAttr()

    @model_validator(mode='before')
    @classmethod
    def validate_root(cls, data):
        """Make sure `data` contains either `poni_file` _or_ `params`,
        not both, and that the field that is used defines a valid
        `pyFAI.azimuthalIntegrator.AzimuthalIntegrator object`.
        """
        if isinstance(data, dict):
            inputdir = data.get('inputdir')
            mask_file = data.get('mask_file')
            params = data.get('params')
            poni_file = data.get('poni_file')
            if mask_file is not None:
                if inputdir is not None and not os.path.isabs(mask_file):
                    data['mask_file'] = mask_file
            if params is not None:
                if poni_file is not None:
                    print('Specify either poni_file or params, not both, '
                          'ignoring poni_file')
                    poni_file = None
            elif poni_file is not None:
                if inputdir is not None and not os.path.isabs(poni_file):
                    data['poni_file'] = poni_file
            else:
                raise ValueError('Specify either poni_file or params')
        return data

    @model_validator(mode='after')
    def validate_ai(self):
        """Set the default azimuthal integrator.

        :return: Validated configuration class.
        :rtype: AzimuthalIntegratorConfig
        """
        if self.params is not None:
            self._ai = AzimuthalIntegrator(**self.params)
        elif self.poni_file is not None:
            # Third party modules
            from pyFAI import load

            self._ai = load(str(self.poni_file))
            self.params = {
                'detector': self._ai.detector.name,
                'dist': self._ai.dist,
                'poni1': self._ai.poni1,
                'poni2': self._ai.poni2,
                'rot1': self._ai.rot1,
                'rot2': self._ai.rot2,
                'rot3': self._ai.rot3,
                'wavelength': self._ai.wavelength,
            }
        return self

    @property
    def ai(self):
        """Return the azimuthal integrator."""
        return self._ai

    @property
    def mask_data(self):
        """Return the mask array to use for this detector from the
        data in the file specified with the `mask_file` field. Return
        `None` if `mask_file` is `None`.
        """
        if self.mask_file is None:
            return None

        import fabio
        _mask_file = fabio.open(self.mask_file)
        mask_data = _mask_file.data
        _mask_file.close()
        return mask_data


class MultiGeometryConfig(CHAPBaseModel):
    """Class representing the configuration for treating simultaneously
    multiple detector configuration within a single integration

    :ivar ais: List of detector IDs of azimuthal integrators.
    :type ais: Union[str, list[str]]
    :ivar azimuth_range: Common azimuthal range for integration,
        defaults to `[-180.0, 180.0]`.
    :type azimuth_range: Union(
        list[float, float], tuple[float, float]), optional
    :ivar radial_range: Common range for integration,
        defaults to `[0.0, 180.0]`.
    :type radial_range: Union(list[float, float],
                              tuple[float, float]), optional
    :ivar unit: Output unit, defaults to `'q_A^-1'`.
    :type unit: str, optional
    """
    ais: conlist(
        min_length=1, item_type=constr(min_length=1, strip_whitespace=True))
    azimuth_range: Optional[
        conlist(
            min_length=2, max_length=2,
            item_type=confloat(ge=-180, le=360, allow_inf_nan=False))
        ] = [-180.0, 180.0]
    radial_range: Optional[
        conlist(
            min_length=2, max_length=2,
            item_type=confloat(ge=0, le=180, allow_inf_nan=False))
        ] = [0.0, 180.0]
    unit: Optional[
        constr(strip_whitespace=True, min_length=1)] = 'q_A^-1'
    chi_disc: Optional[int] = 180
    empty: Optional[confloat(allow_inf_nan=False)] = 0.0
    wavelength: Optional[confloat(allow_inf_nan=False)] = None

    @field_validator('ais', mode='before')
    @classmethod
    def validate_ais(cls, ais):
        """Validate the detector IDs of the azimuthal integrators.

        :param ais: The detector IDs.
        :type ais: str, list[str]
        :return: The detector ais.
        :rtype: list[str]
        """
        if isinstance(ais, str):
            return [ais]
        return ais


class IntegrateConfig(CHAPBaseModel):
    """Class with the input parameters to perform various integrations
    with `pyFAI`.

    :ivar error_model: When the variance is unknown, an error model
        can be given (ignored for radial integration):
        `poisson` (variance = I) or `azimuthal` (variance = (I-<I>)^2).
    :type error_model: str, optional
    """
    # correctSolidAngle: true
    # dark: None
    error_model: Optional[constr(strip_whitespace=True, min_length=1)] = None
    # filename: None
    # flat: None
    # mask: None
    # metadata: None
    # normalization_factor: Optional[confloat(allow_inf_nan=False)] = 1.0
    # polarization_factor: None
    # variance: None
    attrs: Optional[dict] = {}


class Integrate1dConfig(IntegrateConfig):
    """Class with the input parameters to perform 1D azimuthal
    integration with `pyFAI`.

    :ivar method: For pyFAI.azimuthalIntegrator.AzimuthalIntegrator
        a registered integration method or a 3-tuple (splitting,
        algorithm, implementation), defaults to `csr`.
        For pyFAI.multi_geometry.MultiGeometry a registered integration
        method, defaults to `splitpixel`.
    :type method: Union[str, tuple], optional
    :ivar npt: Number of integration points, defaults to 1800.
    :type npt: int, optional
    """
    method: Optional[Union[
        str,
        conlist(
            min_length=3, max_length=3,
            item_type=constr(strip_whitespace=True, min_length=1))]] = None
    npt: Optional[conint(gt=0)] = 1800


class Integrate2dConfig(IntegrateConfig):
    """Class with the input parameters to perform 2D azimuthal (cake)
    integration with `pyFAI`.

    :ivar method: Registered integration method, defaults to `bbox`
        for pyFAI.azimuthalIntegrator.AzimuthalIntegrator or
        `splitpixel` for pyFAI.multi_geometry.MultiGeometry.
    :type method: str, optional
    :ivar npt_azim: Number of points for the integration in the
        azimuthal direction, defaults to 3600.
    :type npt_azim: int, optional
    :ivar npt_rad: Number of points for the integration in the
        radial direction, defaults to 1800.
    :type npt_rad: int, optional
    """
    method: Optional[str] = None
    npt_azim: Optional[conint(gt=0)] = 3600
    npt_rad: Optional[conint(gt=0)] = 1800


class IntegrateRadialConfig(IntegrateConfig, MultiGeometryConfig):
    """Class with the input parameters to perform radial integration
    with `pyFAI`.

    :ivar method: Registered integration method, defaults to `csr`.
    :type method: str, optional
    :ivar radial_unit: Unit used for radial representation,
        defaults to `'q_A^-1'`.
    :type radial_unit: str, optional
    :ivar npt: Number of integration points, defaults to 1800.
    :type npt: int, optional
    """
    radial_unit: Optional[
        constr(strip_whitespace=True, min_length=1)] = 'q_A^-1'
    method: Optional[str] = 'csr'
    npt: Optional[conint(gt=0)] = 1800


class PyfaiIntegratorConfig(CHAPBaseModel):
    """Class representing the configuration for detector data
    integrator for `pyFAI`.

    :ivar right_handed: For radial and cake integration, reverse the
        direction of the azimuthal coordinate from pyFAI's convention,
        defaults to True.
    :type right_handed: bool, optional
    """
    name: constr(strip_whitespace=True, min_length=1)
    integration_method: Literal[
        'integrate1d', 'integrate2d', 'integrate_radial']
    multi_geometry: Optional[MultiGeometryConfig] = None
    integration_params: Optional[
            Union[Integrate1dConfig, Integrate2dConfig, IntegrateRadialConfig]
        ] = None
    right_handed: bool = True

    _placeholder_result: PrivateAttr = None

    @model_validator(mode='before')
    @classmethod
    def validate_config(cls, data):
        """Validate the input integration configuration.

        :param data: Pydantic validator data object.
        :type data: PyfaiIntegratorConfig,
            pydantic_core._pydantic_core.ValidationInfo
        :return: The currently validated list of class properties.
        :rtype: dict
        """
        integration_method = data.get('integration_method')
        integration_params = data.get('integration_params')
        multi_geometry = data.get('multi_geometry')
        if integration_method == 'integrate1d':
            if multi_geometry is None:
                method = 'csr'
            else:
                method = 'splitpixel'
            if integration_params is None:
                data['integration_params'] = Integrate1dConfig(method=method)
            else:
                if integration_params.get('method') is None:
                    integration_params['method'] = method
                data['integration_params'] = Integrate1dConfig(
                    **integration_params)
        elif integration_method == 'integrate2d':
            if multi_geometry is None:
                method = 'bbox'
            else:
                method = 'splitpixel'
            if integration_params is None:
                data['integration_params'] = Integrate2dConfig(method=method)
            else:
                if integration_params.get('method') is None:
                    integration_params['method'] = method
                data['integration_params'] = Integrate2dConfig(
                    **integration_params)
        elif integration_method == 'integrate_radial':
            data['integration_params'] = IntegrateRadialConfig(
                    **integration_params)
        if (integration_method != 'integrate_radial'
                and 'multi_geometry' not in data):
            mg = MultiGeometryConfig(**integration_params)
            data['multi_geometry'] = mg
            data['integration_params'].attrs.update(mg.model_dump(
                include={'azimuth_range', 'radial_range',  'unit'}))

        return data

    @property
    def result_shape(self):
        """Return shape of one frame of results from this integration."""
        if self.integration_method == 'integrate_radial':
            return (self.integration_params.npt, )
        elif self.integration_method == 'integrate1d':
            return (self.integration_params.npt, )
        elif self.integration_method == 'integrate2d':
            return (self.integration_params.npt_azim,
                    self.integration_params.npt_rad)
        else:
            raise NotImplementedError(
                f'Unimplemented integration_method: {self.integration_method}')

    @property
    def result_coords(self):
        """Return a dictionary representing the `zarr.array` objects
        for the coordinates of a single frame of results from this
        integration.
        """
        if self._placeholder_result is None:
            raise RuntimeError('Missing placeholder results')
        return self._placeholder_result['coords']

    def get_axes_indices(self, dataset_ndims):
        """Return the index of each coordinate orienting a single
        frame of results from this integration.
        """
        return {k: dataset_ndims + i
                for i, k in enumerate(self.result_coords.keys())}

    def get_placeholder_data(self, ais):
        """Return empty input data of the correct shape for use in
        `init_placeholder_data`.
        """
        if self.integration_method == 'integrate_radial':
            return {ai:np.full(ais[ai].ai.detector.shape, 0)
                   for ai in self.integration_params.ais}
        return {ai:np.full(ais[ai].ai.detector.shape, 0)
               for ai in self.multi_geometry.ais}

    def init_placeholder_results(self, ais):
        """Get placeholder results for this integration so we can fill
        in the datasets for results of coordinates when setting up a
        zarr tree for holding results of
        `saxswaxs.PyfaiIntegrationProcessor`.
        """
        self._placeholder_result = self.integrate(
            ais, self.get_placeholder_data(ais))

    def integrate(self, azimuthal_integrators, data):
        """Perform the integration and return the results.

        :param azimuthal_integrators: List of single-detector
            integrator configurations.
        :type azimuthal_integrators: list[AzimuthalIntegratorConfig]
        :param data: Dictionary of 2D detector frames to be
            integrated.
        :type data: dict[str, np.ndarray]
        :return: Integrated intensities and coordinates for every
            frame (or set of frames) in `input_data`.
        :rtype: dict[str, object]
        """
        # Third party modules
        from pyFAI.units import (
            AZIMUTHAL_UNITS,
            RADIAL_UNITS,
            to_unit,
        )

        ais = {name: ai.ai for name, ai in azimuthal_integrators.items()}

        if self.integration_method == 'integrate_radial':
            # Third party modules
            from pyFAI.containers import Integrate1dResult

            results = None
            npts = []
            for k, v in data.items():
                if v.ndim == 2:
                    data[k] = np.expand_dims(v, 0)
                elif v.ndim != 3:
                    raise ValueError(
                        f'Illegal dimension for {k} data ({v.ndim})')
                npts.append(data[k].shape[0])
            if not all(n == npts[0] for n in npts):
                raise RuntimeError('Different number of detector frames for '
                                   f'each azimuthal integrator ({npts})')
            npts = npts[0]
            integration_params = self.integration_params.model_dump(
                exclude={'ais', 'attrs', 'error_model', 'chi_disc', 'empty',
                         'wavelength'})
            for name in self.integration_params.ais:
                ai = ais[name]
                mask = azimuthal_integrators[name].mask_data
                _results = [
                    ai.integrate_radial(
                        data=data[name][i],
                        mask=mask,
                        **integration_params)
                    for i in range(npts)
                ]
                if results is None:
                    results = _results
                else:
                    results = [
                        Integrate1dResult(
                            radial=_results[i].radial,
                            intensity=
                                _results[i].intensity + results[i].intensity)
                        for i in range(npts)
                    ]
            if self.right_handed:
                results = [
                    Integrate1dResult(
                        radial=r.radial,
                        intensity=np.flip(r.intensity)
                    )
                    for r in results
                ]
            results = [
                Integrate1dResult(
                    radial=r.radial,
                    intensity=np.where(r.intensity==0, np.nan, r.intensity)
                )
                for r in results
            ]
            # Integrate1dResult's "radial" property is misleadingly
            # named here. When using integrate_radial, the property
            # actually contains azimuthal coordinate values.
            azimuthal_unit = to_unit(
                integration_params['unit'], type_=AZIMUTHAL_UNITS)
            # pyFAI doesn't add the correct unit for chi_deg
            if azimuthal_unit.name[-3:] == 'deg':
                azimuthal_unit.unit_symbol = '$^{o}$'
            coords = {
                azimuthal_unit.name: {
                    'attributes': {
                        'units': azimuthal_unit.unit_symbol,
                        'long_name': azimuthal_unit.label,
                    },
                    'data': results[0].radial,
                    'shape': results[0].radial.shape,
                    'dtype': 'float32',
                },
            }
            results = {'intensities': [v.intensity for v in results],
                       'coords': coords}
        else:
            npts = []
            for k, v in data.items():
                if v.ndim == 2:
                    data[k] = np.expand_dims(v, 0)
                elif v.ndim != 3:
                    raise ValueError(
                        f'Illegal dimension for {k} data ({v.ndim})')
                npts.append(data[k].shape[0])
            if not all(n == npts[0] for n in npts):
                raise RuntimeError('Different number of detector frames for '
                                   f'each azimuthal integrator ({npts})')
            npts = npts[0]
            if self.multi_geometry is None:
                raise RuntimeError('self.multi_geometry is None')
                if len(data) != 1:
                    raise RuntimeError(
                        'Multiple detector not tested without multi_geometry')
                _id = list(ais.keys())[0]
                ai = list(ais.values())[0]
                integration_method = getattr(ai, self.integration_method)
                integration_params = self.integration_params.model_dump()
                integration_params = {
                    **integration_params, **integration_params['attrs']}
                del integration_params['attrs']
                raise RuntimeError(f'Check use of mask in integration_method')
                if masks is None:
                    results = [
                        integration_method(data[_id][i], **integration_params)
                        for i in range(npts)
                    ]
                else:
                    results = [
                        integration_method(
                            np.where(
                                 masks[ai], 0, data[_id][i].astype(np.float64),
                            **integration_params))
                        for i in range(npts)
                    ]
            else:
                # Third party modules
                from pyFAI.multi_geometry import MultiGeometry

                mg = MultiGeometry(
                    [ais[ai] for ai in self.multi_geometry.ais],
                    **self.multi_geometry.model_dump(exclude={'ais'}))
                integration_method = getattr(mg, self.integration_method)
                lst_mask = []
                for name in self.multi_geometry.ais:
                    lst_mask.append(azimuthal_integrators[name].mask_data)
                results = [
                    integration_method(
                        [data[ai][i].astype(np.float64)
                         for ai in self.multi_geometry.ais],
                        lst_mask=lst_mask,
                        **self.integration_params.model_dump(
                            exclude='attrs'))
                    for i in range(npts)
                ]
            if isinstance(self.integration_params, Integrate1dConfig):
                if self.multi_geometry is None:
                    unit = integration_params['unit']
                else:
                    unit = self.multi_geometry.unit
                radial_unit = to_unit(unit, type_=RADIAL_UNITS)
                coords = {
                    radial_unit.name: {
                        'attributes': {
                            'units': radial_unit.unit_symbol,
                            'long_name': radial_unit.label,
                        },
                        'data': results[0].radial,
                        'shape': results[0].radial.shape,
                        'dtype': 'float32',
                    },
                }
                results = {'intensities': [v.intensity for v in results],
                           'coords': coords}
            else:
                # pyFAI doesn't add the correct unit for chi_deg
                azimuthal_unit = deepcopy(results[0].azimuthal_unit)
                if azimuthal_unit.name[-3:] == 'deg':
                    azimuthal_unit.unit_symbol = '$^{o}$'
                coords = {
                    azimuthal_unit.name: {
                        'attributes': {
                            'units': azimuthal_unit.unit_symbol,
                            'long_name': azimuthal_unit.label,
                        },
                        'data': results[0].azimuthal,
                        'shape': results[0].azimuthal.shape,
                        'dtype': 'float32',
                    },
                    results[0].radial_unit.name: {
                        'attributes': {
                            'units': results[0].radial_unit.unit_symbol,
                            'long_name': results[0].radial_unit.label,
                        },
                        'data': results[0].radial,
                        'shape': results[0].radial.shape,
                        'dtype': 'float32',
                    },
                }
                if self.right_handed:
                    intensities = [
                        np.flip(v.intensity, axis=0) for v in results]
                else:
                    intensities = [v.intensity for v in results]
                results = {'intensities': intensities, 'coords': coords}

        return results

    def zarr_tree(self, dataset_shape, dataset_chunks='auto'):
        """Return a dictionary representing a `zarr.group` that can be
        used to contain results from this integration.
        """
        # Third party modules
        import json

        tree = {
            # NXprocess
            'attributes': {
                'default': 'data',
                # 'config': json.dumps(self.dict())
            },
            'children': {
                'data': {
                    # NXdata
                    'attributes': {
                        # 'axes': self.result_axes(),
                        **self.get_axes_indices(len(dataset_shape))
                    },
                    'children': {
                        'I': {
                            # NXfield
                            'attributes': {
                                'long_name': 'Intensity (a.u)',
                                'units': 'a.u'
                            },
                            'dtype': 'float64',
                            'shape': (*dataset_shape, *self.result_shape),
                            'chunks': (*dataset_chunks, *self.result_shape)
                                      if dataset_chunks != 'auto' else 'auto',
                            'compressors': None,
                        },
                        **self.result_coords,
                    }
                }
            }
        }
        return tree


class PyfaiIntegrationConfig(CHAPBaseModel):
    """Class defining components needed for performing one or more
    integrations on the same set of 2D input data with
    `saxswaxs.PyfaiIntegrationProcessor`.
    """
    azimuthal_integrators: Optional[conlist(
        min_length=1, item_type=AzimuthalIntegratorConfig)] = None
    integrations: conlist(min_length=1, item_type=PyfaiIntegratorConfig)
    sum_axes: Optional[bool] = False
    #sum_axes: Optional[
    #    Union[bool, conlist(min_length=1, item_type=str)]] = False

    @model_validator(mode='before')
    @classmethod
    def validate_config(cls, data):
        """Ensure that a valid configuration was provided and finalize
        PONI filepaths.

        :param data: Pydantic validator data object.
        :type data: GiwaxsConversionConfig,
            pydantic_core._pydantic_core.ValidationInfo
        :return: The currently validated list of class properties.
        :rtype: dict
        """
        if isinstance(data, dict):
            inputdir = data.get('inputdir')
            if inputdir is not None and 'azimuthal_integrators' in data:
                ais = data.get('azimuthal_integrators')
                for i, ai in enumerate(deepcopy(ais)):
                    if isinstance(ai, dict):
                        poni_file = ai['poni_file']
                        if not os.path.isabs(poni_file):
                            ais[i]['poni_file'] = os.path.join(
                                inputdir, poni_file)
                    else:
                        poni_file = ai.poni_file
                        if not os.path.isabs(poni_file):
                            ais[i].poni_file = os.path.join(
                                inputdir, poni_file)
                data['azimuthal_integrators'] = ais
        return data

    def zarr_tree(self, dataset_shape, dataset_chunks='auto'):
        """Return a dictionary representing a `zarr.group` that can be
        used to contain results from `saxswaxs.PyfaiIntegrationProcessor`.
        """
        ais = {ai.get_id(): ai for ai in self.azimuthal_integrators}
        for integration in self.integrations:
            integration.init_placeholder_results(ais)
        tree = {
            'root': {
                'attributes': {
                    'description': 'Container for processed SAXS/WAXS data'
                },
                'children': {
                    integration.name: integration.zarr_tree(
                        dataset_shape, dataset_chunks)
                    for integration in self.integrations
                }
            }
        }
        return tree



#!/usr/bin/env python
#-*- coding: utf-8 -*-
"""
File       : processor.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Processor module

Define a generic `Processor` object.
"""

# System modules
import argparse
import logging
from sys import modules
from typing import Optional

# Third party modules
from pydantic import model_validator

# Local modules
from CHAP.pipeline import PipelineItem


class Processor(PipelineItem):
    """Generic data processor.

    The job of any `Processor` in a `Pipeline` is to receive data
    returned by the previous `PipelineItem`, process it in some way,
    and return the result for the next `PipelineItem` to use as input.
    """
    @model_validator(mode='before')
    @classmethod
    def validate_processor_before(cls, data):
        # System modules
        from copy import deepcopy

        # Local modules
        from CHAP.utils.general import (
            dictionary_update,
            is_str_or_str_series,
        )

        if isinstance(data, dict):
            if 'data' in data and 'modelmetaclass' in data:
                mmc = data['modelmetaclass']
                pipeline_fields = mmc.model_fields.get('pipeline_fields')
                if pipeline_fields is not None:
                    for k, v in pipeline_fields.default.items():
                        if is_str_or_str_series(v, log=False):
                            schema = v
                            merge_key_paths = None
                        else:
                            schema = v.get('schema')
                            merge_key_paths = v.get('merge_key_paths')
                        try:
                            value = deepcopy(mmc.get_data(
                                data['data'], schema=schema, remove=False))
                        except:
                            pass
                        else:
                            if k in data:
                                data[k] = dictionary_update(
                                    value, data[k],
                                    merge_key_paths=merge_key_paths,
                                    sort=True)
                            else:
                                data[k] = value
        return data

    def process(self, data):
        """Extract the contents of the input data, add a string to it,
        and return the amended value.

        :param data: Input data.
        :return: Processed data.
        """
        # If needed, extract data from a returned value of Reader.read
        if isinstance(data, list):
            if all(isinstance(d, dict) for d in data):
                data = data[0]['data']
        if data is None:
            return []
        # The process operation is a simple string concatenation
        data += 'process part\n'
        # Return data back to pipeline
        return data


class OptionParser():
    """User based option parser."""
    def __init__(self):
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument(
            '--data', action='store',
            dest='data', default='', help='Input data')
        self.parser.add_argument(
            '--processor', action='store',
            dest='processor', default='Processor', help='Processor class name')
        self.parser.add_argument(
            '--log-level', choices=logging._nameToLevel.keys(),
            dest='log_level', default='INFO', help='logging level')


def main(opt_parser=OptionParser):
    """Main function."""
    optmgr = opt_parser()
    opts = optmgr.parser.parse_args()
    cls_name = opts.processor
    try:
        processor_cls = getattr(modules[__name__], cls_name)
    except AttributeError:
        print(f'Unsupported processor {cls_name}')
        raise

    processor = processor_cls()
    processor.logger.setLevel(getattr(logging, opts.log_level))
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter(
        '{name:20}: {message}', style='{'))
    processor.logger.addHandler(log_handler)
    processor.process(opts.data)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#-*- coding: utf-8 -*-
"""
File       : utils.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: common set of utility functions
"""

# System modules
import os
import sys
import platform
import subprocess
import json

# Third party modules
from importlib.metadata import distributions

def osinfo():
    """Helper function to provide osinfo."""
    os_info = {
        'name': f'{platform.system().lower()}-{platform.release()}',
        'kernel': platform.version(),
        'version': platform.platform(),
    }
    return os_info

def environments():
    """Detects the current Python environment (system, virtualenv,
    Conda, or pip) and collects package information. Returns a list
    of detected environments with installed packages.
    """
    environments_ = []
    os_name = f'{platform.system().lower()}-{platform.release()}'

    # Check for Conda environment
    conda_env = os.getenv('CONDA_PREFIX')
    if conda_env:
        conda_env_name = os.getenv('CONDA_DEFAULT_ENV', 'unknown-conda-env')
        try:
            # Fetch Conda packages
            conda_packages = subprocess.check_output(
                ['conda', 'list', '--json'], text=True)
            conda_packages = json.loads(conda_packages)
            packages = [{'name': pkg['name'], 'version': pkg['version']}
                        for pkg in conda_packages]
        except Exception:
            packages = []
        environments_.append({
            'name': conda_env_name,
            'version': sys.version.split()[0],
            'details': 'Conda environment',
            'parent_environment': None,
            'os_name': os_name,
            'packages': packages,
        })

    # Check for Virtualenv (excluding Conda)
    elif hasattr(sys, 'real_prefix') or os.getenv('VIRTUAL_ENV'):
        venv_name = os.path.basename(os.getenv('VIRTUAL_ENV', 'unknown-venv'))
        packages = [{'name': pkg.metadata['Name'],
                     'version': pkg.version}
                    for pkg in distributions()]
        environments_.append({
            'name': venv_name,
            'version': sys.version.split()[0],
            'details': 'Virtualenv environment',
            'parent_environment': None,
            'os_name': os_name,
            'packages': packages,
        })

    # System Python (not inside Conda or Virtualenv)
    else:
        packages = [{'name': pkg.metadata['Name'],
                     'version': pkg.version}
                    for pkg in distributions()]
        environments_.append({
            'name': 'system-python',
            'version': sys.version.split()[0],
            'details': 'System-wide Python',
            'parent_environment': None,
            'os_name': os_name,
            'packages': packages,
        })

    # Check for PIP installed editable packages
    try:
        # Fetch Git repo info for PIP installed editable packages
        pip_packages = subprocess.check_output(
            ['pip', 'list', '--format', 'json'], text=True)
        pip_packages = json.loads(pip_packages)
        packages = [{'name': pkg['name'],
                     #'version': pkg['version'],
                     'commit_hash': subprocess.check_output(
                         ["git", "rev-parse", "HEAD"],
                         cwd=pkg['editable_project_location']
                     ).strip().decode()}
                    for pkg in pip_packages
                    if 'editable_project_location' in pkg]
        environments_.append({
            'name': 'Python Package Installer',
            'version': sys.version.split()[0],
            'details': 'Editable project locations',
            'parent_environment': None,
            'os_name': os_name,
            'packages': packages,
        })
    except Exception:
        pass

    return environments_
